;; $Id: blocker-ltk.lisp,v 1.22 2005/05/21 22:10:23 jimka Exp $

#|

 This software is Copyright (c) 2005 Jim Newton <jimka@rdrop.com>

 Jim Newton grants you the rights to distribute
 and use this software as governed by the terms
 of the Lisp Lesser GNU Public License
 (http://opensource.franz.com/preamble.html),
 known as the LLGPL.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
|#

(in-package blocker)

(defclass ltk-blocker ( blocker-base)
  (( view-point
     :accessor view-point)
   ( blocker
     :accessor blocker
     :initarg :blocker)
   ( virtual
     :accessor virtual
     :initarg :virtual
     :type virtual-3d-equipment)
   ( color-mapping
     :initform (make-hash-table)
     :accessor color-mapping)
   ( alpha
     :accessor alpha
     :initform (* 45/180 pi )
     )
   ( beta
     :accessor beta
     :initform (* 150/180 pi )
     )
   ( theta
     :accessor theta
     :initform (* 30/180 pi)
     )

   ( min-x
     :initform 0
     :accessor min-x)
   ( min-y
     :accessor min-y
     :initform 0)
   ( max-x
     :accessor max-x
     :initform 1)
   ( max-y
     :accessor max-y
     :initform 1)
   
   ( scale-x
     :initform 30
     :accessor scale-x)
   ( scale-y
     :initform -30
     :accessor scale-y)
   ( origin-x
     :accessor origin-x
     :initform 0)
   ( origin-y
     :accessor origin-y
     :initform -50)
   
   ( view-perspecitve-matrix
     :accessor view-perspecitve-matrix
     :type matrix)
   ( canvas
     :accessor canvas)
   ( board
     :accessor board)
   ( tk-faces
     :accessor tk-faces
     :initform nil)))

(defun report-event (event)
  (fresh-line)
  (format t "event=~A~%" event)
  (finish-output))


(defmethod initialize-instance :after (( gui ltk-blocker) &rest ignored)
  (declare (ignore ignored))
  (format t "initializing ltk-blocker~%")
  (setf ltk:*debug-tk* nil)
  (setf ltk:*wish-args* '("-name" "Blocker (ltk)"
			  "-geometry" "400x400"))
  (ltk:start-wish)

  (let* ((canvas (make-instance 'ltk:canvas))
	 (alpha  (make-instance 'ltk:entry :master canvas))
	 (beta   (make-instance 'ltk:entry :master canvas))
	 (theta  (make-instance 'ltk:entry :master canvas)))
    (setf (ltk:text alpha) (alpha gui))
    (setf (ltk:text beta) (beta gui))
    (setf (ltk:text theta) (theta gui))
    (ltk:pack canvas :side :top)
    (ltk:pack alpha :side :left)
    (ltk:pack beta :side :left)
    (ltk:pack theta :side :left)
    (ltk:bind canvas
	      "<Leave>"
	      (lambda (event)
		(report-event event)
		(setf (alpha gui) (ltk:text alpha)
		      (beta gui)  (ltk:text beta)
		      (theta gui) (ltk:text theta))
		(format t "alpha=~A beta=~A theta=~A~%"
			(alpha gui)
			(beta gui)
			(theta gui)))))
  
  (setf (view-perspecitve-matrix gui)
	(rotate-3-angle (alpha gui) (beta gui) (theta gui))

	(view-point gui)
	(apply-transform (scale-point 0.5 (make-3d-point 0
						    (num-cols (blocker gui))
						    50))
		    (unrotate-3-angle (alpha gui) (beta gui) (theta gui)))
	)

  (let* (( sc (make-instance 'ltk:scrolled-canvas))
	 ( canvas (ltk:canvas sc))
	 ( color-mapping (color-mapping gui)))
    
    (setf (gethash 0 color-mapping) :yellow
	  (gethash 1 color-mapping) :blue
	  (gethash 2 color-mapping) :white
	  (gethash 3 color-mapping) :black
	  (gethash 4 color-mapping) :blue
	  (gethash 5 color-mapping) :orange)
    
    (setf (canvas gui) canvas)
    (assert (canvas gui))
    (ltk:pack sc :expand 1 :fill :both)

    (initialize-board gui)))


(defmethod tk-color (( gui ltk-blocker) ( color integer))
  (or (gethash color (color-mapping gui))
      (error "no color mapping for ~A~%" color)))

(defclass tk-face ( blocker-base )
  (( face
     :accessor face
     :initarg :face)
   ( polygon
     :accessor polygon
     :initarg :polygon)))   

(defmethod center (( tk-face tk-face))
  (center (face tk-face)))    

(defmethod create-colored-polygon (( gui ltk-blocker) (face face))
  ;; create-polygon takes a list of points of the form ( x0 y1 x1 y1 x2 y2 ...)
  (let (( poly (make-instance 'ltk:canvas-polygon
			      :canvas (canvas gui)
			      :coords (pt-to-canvas gui face))))
    (ltk:configure poly
		   :fill (tk-color gui (color face))
		   :outline :orange)
    poly))
  

;; returns a ltk-face object
(defmethod draw-face (( gui ltk-blocker) ( face face))
  (make-instance 'tk-face
		 :face face
		 :polygon (create-colored-polygon gui face)))

;; returns a list of the form ( x0 y1 x1 y1 x2 y2 ...)
(defmethod pt-to-canvas (( gui ltk-blocker) ( face face))
  (loop for pt across (points face)
	nconcing (pt-to-canvas gui pt)))

;; returns a list of the form ( x0 y1 x1 y1 x2 y2 ...)
(defmethod pt-to-canvas (( gui ltk-blocker) ( pts list))
  (loop for pt in pts
	nconcing (pt-to-canvas gui pt)))

(defmethod mid-point (( face1 face) ( face2 face))
  (mid-point (center face1) (center face2)))

(defmethod pt-to-canvas (( gui ltk-blocker) ( face polyhedran-face))
  (assert (opposite-face face))
  (let (( mid (mid-point face (opposite-face face)))
	( epsilon 0.08)
	pt_calc
	)
    (loop for pt across (points face)
	  ;; calculate a point between the pt and mid
	  nconcing (progn
		     (setf pt_calc
			   (add-point pt
				      (scale-point
				       (make-3d-point (signum (- (x-coord mid) (x-coord pt)))
						      (signum (- (y-coord mid) (y-coord pt)))
						      (signum (- (z-coord mid) (z-coord pt))))
				       epsilon)))
		     (setf pt_calc
			   (add-point pt_calc
				      (make-3d-point 0 0 epsilon)))
		     (pt-to-canvas gui pt_calc)))))    

(defmethod initialize-board (( gui ltk-blocker))
  ;; draw grid parts
  ;;(assert (virtual gui))
  ;;(assert (board-grid (virtual gui)))
  ;;(assert (faces (board-grid (virtual gui))))

  (loop for face across (faces (board-grid (virtual gui)))
	collect (draw-face gui face))

  ;; draw stone parts
  (setf (tk-faces gui) (loop for vp in (virtual-pieces (virtual gui))
			     nconcing
			     (loop for face across (faces (polyhedran vp))
				   collect (tk-bind-face gui face (piece vp)))))
  
  (sort-pieces gui)
  (scroll-to-view gui))

  

(defmethod tk-bind-face (( gui ltk-blocker) (pf polyhedran-face) (piece piece))
  (let* ((face  (draw-face gui pf))
	 (canvas (canvas gui))
	 (poly  (polygon face)))
    (ltk:itembind canvas
		  poly
		  "<Enter>"
		  (lambda (event)
		    (report-event event)
		    (ltk:configure poly :fill :green)))
    (ltk:itembind canvas
		  poly
		  "<Leave>"
		  (lambda (event)
		    (report-event event)
		    (ltk:configure poly :fill (color face))))
    (dolist (key '("n" "s" "e" "w"))
      (ltk:itembind canvas
		    poly
		    key
		    (lambda (event)
		      (report-event event)
		      (ltk:configure poly :fill (color face))
		      (tk-select-event piece key))))
    
    face ))

(defmethod update-board (( gui ltk-blocker))
  (sort-pieces gui)
  (loop for tk-face in (tk-faces gui)
	do 
	(ltk:set-coords (canvas gui) (polygon tk-face)
			(pt-to-canvas gui (face tk-face))))

  ;; (loop for pt in (list (make-3d-point 0 0 0)
  ;; (make-3d-point 0 11 0)
  ;; (make-3d-point 11 11 0)
  ;; (make-3d-point 5.5 5.5 0)
  ;; (make-3d-point 11 0 0))
  ;; do
  ;; (make-instance 'ltk:canvas-line
  ;; :canvas (canvas gui)
  ;; :coords (pt-to-canvas gui (list (view-point gui) pt))))

  (scroll-to-view gui))

(defmethod scroll-to-view (( gui ltk-blocker))
  (ltk:scrollregion (canvas gui) 
		    (round (min-x gui))
		    (round (min-y gui))
		    (round (max-x gui))
		    (round (max-y gui))))

(defmethod dist (( pt point-3d) ( tk-face tk-face))
  (dist pt (face tk-face)))

;(defmethod dist (( pt point-3d) ( face face))
;  (loop for vertex across (points face)
;	maximizing (dist pt vertex)))

(defmethod dist (( pt point-3d) ( face face))
  (loop for vertex across (points face)
	sum (dist pt vertex)))

(defmethod sort-pieces (( gui ltk-blocker))
  (let ((last nil))

    (loop for tk-face in (sort (copy-list (tk-faces gui))
			       #'(lambda ( tk-face-a tk-face-b)
				   (> (dist (view-point gui) tk-face-a)
				      (dist (view-point gui) tk-face-b))))
			      
	  do (progn
	       (when last
		 (ltk:raise (polygon tk-face) last))
	       (setf last (polygon tk-face)))
	  )))


;; returns a 2d point in the vector space of the tk canvas
;; given a point in the 3d vector space of the blocker game board
(defmethod pt-to-canvas (( gui ltk-blocker) ( pt point-3d))
  (let* (( new-pt (apply-transform pt (view-perspecitve-matrix gui)))
	 ( x (+ (origin-x gui) (* (scale-x gui) (x-coord new-pt))))
	 ( y (+ (origin-y gui) (* (scale-y gui) (y-coord new-pt)))))

    (setf (min-x gui) (min x (min-x gui))
	  (min-y gui) (min y (min-y gui))
	  (max-x gui) (max x (max-x gui))
	  (max-y gui) (max y (max-y gui)))
    (list (round x) (round y))))
	  

(defmethod handle-event (( gui ltk-blocker) ( ev move-direction))
  (update-board gui))

(define-condition break-out-of-mainloop ()
  (( event-string :initarg :event-string :reader event-string)))
(defmethod read-input (( self ltk:widget))
  (handler-case
      (progn
	(fresh-line)
	(format t "Waiting on mouse event~%")
	(finish-output)
	(ltk:mainloop))
    (break-out-of-mainloop (se)
	(return-from read-input (event-string se)))))

(defmethod tk-select-event (( piece piece) (dir string))
  (signal (make-condition 'break-out-of-mainloop
			  :event-string (format nil "~A~A"
						(printable piece)
						dir))))

(defmethod tk-becomes (( player player))
  (setf (input-file player)
	ltk:*tk*))


(defclass tk-becomes-white (blocker-event)
  (( name
     :initform "TK GUI White")
   ( shortcut
     :initform "tw")))

(defmethod handle-event (( blo blocker) ( self tk-becomes-white))
  (tk-becomes (white blo)))

(defmethod legal-p (( self tk-becomes-white))
  (not (eql ltk:*tk* (input-file (white (blocker self))))))

(defclass tk-becomes-black (blocker-event)
  (( name
     :initform "TK GUI Black")
   ( shortcut
     :initform "tb")))

(defmethod handle-event (( blo blocker) ( self tk-becomes-black))
  (tk-becomes (black blo)))

(defmethod legal-p (( self tk-becomes-black))
  (not (eql ltk:*tk* (input-file (black (blocker self))))))


(defmethod event-setup (( gui ltk-blocker))
  (declare (ignore gui))
  (format t "~%~%~%~%handling excessive TK events~%~%~%~%")
  (ltk:handle-pending-events))
