;; $Id: event.lisp,v 1.6 2005/05/16 12:11:37 jimka Exp $

#|

 This software is Copyright (c) 2005 Jim Newton <jimka@rdrop.com>

 Jim Newton grants you the rights to distribute
 and use this software as governed by the terms
 of the Lisp Lesser GNU Public License
 (http://opensource.franz.com/preamble.html),
 known as the LLGPL.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
|#

(in-package blocker)

(defmethod prompt (( self blocker-event))
  (name self))

(defmethod printable (( self blocker-event))
  (prompt self))

(defmethod legal-p (( self blocker-event))
  (declare (ignore self))
  t)

(defmethod handle-event (( blo blocker) ( self blocker-event))
  (declare (ignore self blo ))
  nil)



(defclass no-event ( blocker-event )
  (( name
     :initform "Return")))

(defclass quit ( blocker-event )
  (( name
     :initform "Quit")
   ( shortcut
     :initform "q!")))

(defun quit-blocker ()
  (throw 'finish-blocker nil))
  

(defmethod handle-event (( blo blocker) ( self quit))
  (declare (ignore blo self))
  (quit-blocker))


(defclass move-piece ( blocker-event )
  (( volume
     :accessor volume)))

(defmethod initialize-instance :after (( self move-piece) &rest ignored)
  (declare (ignore ignored))
  (setf (shortcut self) (format nil "~A" (volume self))
	(name self) (shortcut self)))

(defmethod handle-event (( blo blocker) ( self move-piece))
  (init-selected-combination (turn blo))
  (let (( piece (find-piece (turn blo) (volume self))))
    (select-piece (turn blo) piece)))

(defmethod prompt (( self move-piece))
  (dolist ( piece (pieces (turn (blocker self))))
    (when (= (volume piece) (volume self))
      (return-from prompt (format nil "~A ~A" (volume piece) (name piece)))))
  nil)

(defmethod legal-p (( self move-piece))

  (cond
    ((selected-piece (turn (blocker self)))
     nil)

    (t
     (any-valid-comb (find-piece (turn (blocker self)) (volume self))))))

(defclass move-0 ( move-piece)
  (( volume
     :initform 0)))

(defclass move-1 ( move-piece)
  (( volume
     :initform 1)))

(defclass move-2 ( move-piece)
  (( volume
     :initform 2)))

(defclass move-4 ( move-piece)
  (( volume
     :initform 4)))

(defclass move-8 ( move-piece)
  (( volume
     :initform 8)))

(defclass move-direction  ( blocker-event )
  nil)

(defmethod initialize-instance :after (( self move-direction) &rest ignored)
  (declare (ignore ignored))
  (setf (name self) (shortcut self)))

(defmethod prompt (( self move-direction))
  (dolist ( dir (directions (blocker self)))
    (when (equal (shortcut self) (-symbol dir))
      (return-from prompt (name dir))))
  nil)

(defmethod legal-p (( self move-direction))
  (let (( piece (selected-piece (turn (blocker self))))
	dir
	)
    (cond
      ((null piece)
       nil)
      
      (t
       (setf dir (find-direction (blocker self) (shortcut self)))
       (is-legal-direction piece dir (selected-combination (turn (blocker self))))))))

(defmethod handle-event (( blo blocker) ( self move-direction))
  (let (( dir (find-direction blo (shortcut self)))
	piece
	)
    (remember-direction (turn blo) dir)
    (setf piece (selected-piece (turn blo)))
    (move-piece dir piece)
    (when (= (length (selected-combination (turn blo)))
	     (number-of-rolls piece))
      ;; unselect-piece?
      (put-event blo "ot"))))
  
(defclass move-s ( move-direction)
  (( shortcut
     :initform "s")))

(defclass move-n ( move-direction)
  (( shortcut
     :initform "n")))

(defclass move-e ( move-direction)
  (( shortcut
     :initform "e")))

(defclass move-w ( move-direction)
  (( shortcut
     :initform "w")))

(defclass move-sw ( move-direction)
  (( shortcut
     :initform "sw")))

(defclass move-nw ( move-direction)
  (( shortcut
     :initform "nw")))

(defclass move-se ( move-direction)
  (( shortcut
     :initform "se")))

(defclass move-ne ( move-direction)
  (( shortcut
     :initform "ne")))

(defclass move-i ( move-direction)
  (( shortcut
     :initform "i")))

(defclass draw-board ( blocker-event )
  (( name
     :initform "Draw Board")
   ( shortcut
     :initform "d")))

(defmethod handle-event (( blo blocker) ( self draw-board))
  (draw blo))

(defclass verbose-yes ( blocker-event)
  (( name
     :initform "Verbose ON")
   ( shortcut
     :initform "v")))

(defmethod legal-p (( self verbose-yes))
  (null (verbose (blocker self))))

(defmethod handle-event (( blo blocker) ( self verbose-yes))
  (verbose blo))

(defclass verbose-no ( blocker-event)
  (( name
     :initform "Verbose OFF")
   ( shortcut
     :initform "v")))

(defmethod legal-p (( self verbose-no))
  (verbose (blocker self)))

(defmethod handle-event (( blo blocker) ( self verbose-no))
  (null (verbose blo)))

(defclass undo ( blocker-event )
  (( name
     :initform "Undo")
   ( shortcut
     :initform "u")))

(defmethod legal-p (( self undo))
  (history (blocker self)))

(defmethod handle-event (( blo blocker) ( self undo))
  (undo-combination blo))

(defclass suggest ( blocker-event )
  (( name
     :initform "Suggest move")
   ( shortcut
     :initform "su")))

(defmethod handle-event (( blo blocker) ( self suggest))
  (format t "SUGGEST: ~A" (printable (calc-best-move (turn blo)))))
    


(defclass human-becomes-white ( blocker-event)
  (( name
     :initform "Human White")
   ( shortcut
     :initform "hw")))

(defmethod handle-event (( blo blocker) ( self human-becomes-white))
  (setf (computer blo)
	(remove (white blo) (computer blo))))

(defmethod legal-p (( self human-becomes-white))
  (null (legal-p (make-instance 'computer-becomes-white :blocker (blocker self)))))

(defclass human-becomes-black ( blocker-event )
  (( name
     :initform "Human Black")
   ( shortcut
     :initform "hb")))

(defmethod handle-event (( blo blocker) ( self human-becomes-black))
  (setf (computer blo)
	(remove (black blo) (computer blo))))

(defmethod legal-p (( self human-becomes-black))
  (null (legal-p (make-instance 'computer-becomes-black :blocker (blocker self)))))


(defclass computer-becomes-white ( blocker-event )
  (( name
     :initform "Computer White")
   ( shortcut
     :initform "cw")))

(defmethod handle-event (( blo blocker) ( self computer-becomes-white))
  (computer-becomes (white blo)))

(defmethod legal-p (( self computer-becomes-white))
  (null (member (white (blocker self)) (computer (blocker self)))))

(defclass computer-becomes-black ( blocker-event)
  (( name
     :initform "Computer Black")
   ( shortcut
     :initform "cb")))

(defmethod handle-event (( blo blocker) ( self computer-becomes-black))
  (computer-becomes (black blo)))

(defmethod legal-p (( self computer-becomes-black))
  (null (member (black (blocker self)) (computer (blocker self)))))

(defclass other-turn ( blocker-event )
  (( name :initform "Other Turn")
   ( shortcut :initform "ot")))

(defmethod legal-p (( self other-turn ))
  (let (( piece (selected-piece (turn (blocker self)))))
    (when piece
      (= (number-of-rolls piece)
	 (length (selected-combination (turn (blocker self))))))))

(defmethod handle-event (( blo blocker) ( self other-turn))
  (deselect-piece (turn blo))
  (init-selected-combination (turn blo))
  (other-turn blo))

(defclass move-automatic ( blocker-event )
  (( name :initform "Move Automatic")
   ( shortcut :initform "ma")))

(defmethod legal-p (( self move-automatic ))
  (member (turn (blocker self)) (computer (blocker self))))

(defmethod handle-event (( blo blocker) ( self move-automatic ))
  (let* (( blocker blo)
	 ( q (event-queue blocker))
	 ( mp (calc-best-move (turn blocker)))
	 ( shortcut (format nil "~A" (volume (piece mp)))))

    (qpush q (find-event-by-shortcut blocker shortcut))
    (dolist ( dir (directions (combination mp)))
      (qpush q (find-event-by-shortcut blocker (-symbol dir))))))
    
(defclass auto-play ( blocker-event )
  (( name :initform "Auto Play")
   ( shortcut :initform "ap")))

(defmethod handle-event (( blo blocker) ( self auto-play ))
  (auto-play blo))


(defclass read-strategy ( blocker-event )
  (( name :initform "Read Strategy Components")
   ( shortcut :initform "rs")))

(defmethod handle-event (( blo blocker) ( self read-strategy))
  (-read-from-file (strategy (black blo))))


(defclass write-strategy ( blocker-event )
  (( name :initform "Write Strategy Components")
   ( shortcut :initform "ws")))

(defmethod handle-event (( blo blocker) ( self write-strategy))
  (-write-to-file (strategy (black blo))))

(defclass valid-combinations ( blocker-event)
  (( name :initform "Show Valid Combinations")
   ( shortcut :initform "vc")))

(defmethod legal-p (( self valid-combinations))
  (and (selected-piece (turn (blocker self)))
       (null (selected-combination (turn (blocker self))))))

(defmethod handle-event (( blo blocker) ( self valid-combinations))
  (write-output blo (format nil "Valid combinations are:~%"))  
  (dolist ( comb (combinations (selected-piece (turn blo))))
    (when (is-valid-comb (turn blo) (selected-piece (turn blo)) comb :verbose nil)
      (write-output blo (format nil "   ~A~%" comb)))))

(defclass valid-events ( blocker-event)
  (( name :initform "Show Valid Events")
   ( shortcut :initform "?")))

(defmethod handle-event (( blo blocker) ( self valid-events))
  (write-output blo "Valid selections are:")
  (dolist ( ev (events blo))
    (when (legal-p ev)
      (write-output blo (format nil "    ~2A: ~A~%" (shortcut ev)
			  (prompt ev))))))

(defclass game-over ( blocker-event)
  (( name :initform "Game Over")
   ( shortcut :initform "q")))

(defmethod handle-event (( blo blocker) (self game-over))
  (setf (game-over blo) t))
