;; $Id: main.lisp,v 1.26 2005/05/21 20:22:36 jimka Exp $

#|

 This software is Copyright (c) 2005 Jim Newton <jimka@rdrop.com>

 Jim Newton grants you the rights to distribute
 and use this software as governed by the terms
 of the Lisp Lesser GNU Public License
 (http://opensource.franz.com/preamble.html),
 known as the LLGPL.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
|#

(defun l nil (load "main.lisp"))
(setf *compile-print* t
      *compile-verbose* t)
(defparameter *gc-verbose* nil)

(defpackage "BLOCKER"
  (:use "COMMON-LISP"
	#+cmu   "PCL"
	#+clisp "CLOS")
  (:export
   "VISUAL-GAME"
   "PLAY-GAME"
   "TK-GAME"
   "SELF-PLAY"))

(in-package blocker)


;; TODO load-compiled-file should return nil if there is a compilation error
(defun load-compiled-file ( name )
  (let ( source-name compiled-name)
    (cond
      ((probe-file (setf source-name (format nil "~A.lisp" name))))
      (t
       (error "cannot file file ~A~%" name)))

    (setf compiled-name (format nil "~A.x86f" name))

    (when (or (null (probe-file compiled-name))
	      (> (file-write-date source-name)
		 (file-write-date compiled-name)))
      (format t "compiling ~A~%" name)
      (compile-file name))

    (load name)))

(defparameter *files* '(
		 "ltk"
		 "ltk-mw"

		 "generic"

		 "blocker"
		 "piece"
		 "game"
		 "event"
		 "strat"
		 "polyhedra"

		 "blocker-ltk"

		 "ui"
			))

;; find and report functions which have methods but no
;; defgeneric defined.
(defun find-missing-defgenerics ()
  (let ((defgenerics (make-hash-table :test #'eql))
	(defmethods  (make-hash-table :test #'eql))
	form)
    (dolist (file *files*)
      (setf file (format nil "./~A.lisp" file))
      (with-open-file (stream file)
	(loop while (setf form (read stream nil nil))
	      do
	      (register-funs form defgenerics defmethods))))
    (loop for fun being the hash-keys of defmethods
	  unless (gethash fun defgenerics)
	  do
	  (format t "   (defgeneric ~A (" fun)
	  (dolist (param (gethash fun defmethods))
	    (format t "~A " (cond
			     ((symbolp param)
			      param)
			     ((listp param)
			      (car param)))))
	  (format t "))~%");


	  )))

(defun register-funs (form defgenerics defmethods)
  (cond
    ((null form) nil)
    ((atom form) nil)
    ((eql 'defgeneric (car form))
     (setf (gethash (cadr form) defgenerics) t))
    ((eql 'defmethod (car form))
     (let ((fun (cadr form)))
       (pop form) ;; trash defmethod
       (pop form) ;; trash function name
       (loop while (and form (symbolp (car form)))
	     do
	     (pop form))
       (setf (gethash fun defmethods) (car form))))
    (t
     (register-funs (car form) defgenerics defmethods)
     (register-funs (cdr form) defgenerics defmethods))))

(defun compile-blocker ()
  (mapc #'load-compiled-file *files*)
  ;; (find-missing-defgenerics)
 )


(compile-blocker)
