;; $Id: strat.lisp,v 1.5 2005/05/23 21:16:20 jimka Exp $

#|

 This software is Copyright (c) 2005 Jim Newton <jimka@rdrop.com>

 Jim Newton grants you the rights to distribute
 and use this software as governed by the terms
 of the Lisp Lesser GNU Public License
 (http://opensource.franz.com/preamble.html),
 known as the LLGPL.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
|#

(in-package blocker)

(defmethod blocker (( self strategy))
  (blocker (player self)))
 
(defmethod player (( self strategy-component))
  (player (strategy self)))

(defmethod component-value (( self strategy-component))
  1)

(defmethod rating (( self strategy-component))
  (let (( value (coerce (component-value self) 'single-float)))
    (+ (* value (car (coef self)))
       (* value value (cadr (coef self))))))


(defmethod strat-to-hash (( self strategy))
  (let (( hash (make-hash-table)))
    (dolist ( player (players (blocker self)))
      (dolist ( comp (components self))
	(setf (gethash (list (name player) (name comp)) hash)
	      (coef comp))))
    hash))

;; modifies the coefs of all the componets of the given strategy
;; object to those found in the hash table
(defmethod hash-to-strat (( self strategy) ( hash hash-table))
  (dolist ( player (players (blocker self)))
    (dolist ( comp (components self))
      (when (gethash (list (name player) (name comp)) hash)
	(setf (coef comp)
	      (gethash (list (name player) (name comp)) hash))))))

(defmethod file-name (( self strategy))
  (format nil "~A/~A" (strategy-dir self) (strategy-base self)))

(defmethod -write-to-file (( self strategy))
  (let (( hash (strat-to-hash self)))
    (with-open-file ( fp (file-name self)
			 :if-does-not-exist :create
			 :if-exists :rename
			 :direction :output)
      (loop for player-comp being the hash-keys of hash
	    do
	    (print (list player-comp (gethash player-comp hash))
		   fp)))))

(defmethod -read-from-file (( self strategy))
  (let (( hash (strat-to-hash self))
	( fp (open (file-name self)
		   :direction :input
		   :if-does-not-exist nil))
		 
	line)
    (when fp
      (loop while (setf line (read fp nil))
	    do
	    (setf (gethash (car line) hash) (cadr line)))
      (close fp)
      (hash-to-strat self hash))))


(defmethod calc-position-rating (( self strategy))
  (loop for comp in (components self)
	summing (rating comp)))


;; VARIOUS STRATEGY COMPONENTS

;; random

(defclass rand ( strategy-component )
  (( name
     :initform "Random" )
   ( coef
     :initform '( 1 2))))

(defmethod component-value (( self rand))
  (random 1.0))


;; distance of pice to godfather


(defclass piece-gf-dist ( strategy-component )
  (( nominal-dist-to-gf
     :accessor nominal-dist-to-gf
     :initform 20.0)
   ( target-volume
     :accessor target-volume)))

(defmethod component-value (( self piece-gf-dist))
  (let* (( player (player self))
	 ( target-piece (find-piece player (target-volume self)))
	 ( gf (godfather (opponent player)))
	 ( closeness (- (nominal-dist-to-gf self)
			(dist target-piece gf))))
    (if (plusp closeness)
	(/ closeness (nominal-dist-to-gf self))
	0)))

   
(defclass frecher-gf-dist ( piece-gf-dist )
  (( name
     :initform "Dist 1 to gf")
   ( target-volume
     :initform 1)
   ( coef
     :initform '( 95.38993 51.33060))))

(defclass thin-gf-dist( piece-gf-dist)
  (( name
     :initform "Dist 2 to gf")
   ( target-volume
     :initform 2)
   ( coef
     :initform '( 36.006981 23.06349))))

(defclass flat-gf-dist ( piece-gf-dist)
  (( name
     :initform  "Dist 4 to gf")
   ( target-volume
     :initform 4)
   ( coef
     :initform '(22.74877 16.78349))))
     

(defclass fat-gf-dist ( piece-gf-dist )
  (( name
     :initform "Dist 8 to gf")
   ( target-volume
     :initform 8)
   ( coef
     :initform '(59.832729  2.980000))))

(defclass player-shadow ( strategy-component )
  (( name
     :initform "Shadow")
   ( coef
     :initform '( 24.54756 2.30000))))

(defmethod component-value (( self player-shadow))
  (/ (shadow-size (player self))
     (size (blocker (player self)))))

(defclass wasted-shadow ( strategy-component)
  (( name
     :initform "Wasted Shadow")
   ( coef
     :initform '( 1.0 1.0))))

(defmethod component-value (( self wasted-shadow ))
  (let (( size (size (blocker (player (strategy self))))))
    (/ (- size (wasted-shadow (player (strategy self)))) size)))
    

(defclass opp-shadow ( strategy-component)
  (( name
     :initform "Opponent Shadow")
   ( coef
     :initform '( 14.17528 24.18000))))

(defmethod component-value (( self opp-shadow))
  (/ (shadow-size (opponent (player self)))
     (size (blocker (player self)))))

(defclass check ( strategy-component)
  (( name
     :initform "Check")
   ( coef
     :initform '( 35.23989 62.443876))))

(defmethod component-value (( self check))
  (let (( player (player self)))
    (/ (loop for piece in (rollable-pieces player)
	   when (is-attacking piece (godfather (opponent player)))
	   count piece)
       4.0)))
		     
		      
(defclass encroachment ( strategy-component)
  (( name
     :initform  "Encroachment of gf")
   ( coef
     :initform '( 13.689633 17.68475))))

(defmethod component-value (( self encroachment))
  (let* (( player (player self))
	 ( opponent (opponent player))
	 ( gf-shadow (get-shadow (godfather opponent)))
	 ( attacked 0 )
	 )

    (when (= 1 (hash-table-count gf-shadow))
      (return-from component-value 1.0))

    (setf attacked
	  (loop for piece in (pieces player)
		summing
		(loop for square being the hash-keys of (get-shadow piece)
		      if (gethash square gf-shadow)
		      count square)))
    (cond
      ((zerop attacked)
       0)
      
      ((plusp (hash-table-count gf-shadow))
       (/ attacked (hash-table-count gf-shadow)))

      (t
       1.0))))
      

(defclass back-line ( strategy-component)
  (( name
     :initform "Back Line")
   ( coef
     :initform '( 19.52000 27.09604))))

(defmethod component-value (( self back-line))
  (/ (closeness-to-backline (godfather (player self)))
     10.0))

(defclass danger-of-check ( strategy-component)
  (( name
     :initform "Danger of Check")
   ( coef
     :initform '( 10.0  80.0))))

(defmethod component-value (( self danger-of-check ))
  (let* (( player (player  self))
	 ( godfather (godfather player))
	 ( shadow (get-shadow godfather))
	 
	 ;; do not count the current position of the godfather in
	 ;;   the shadow
	 ( square (get-square (blocker godfather)
			      (y-coord (state godfather))
			      (x-coord (state godfather))))
	 
	 ( shadow-size (hash-table-count shadow)))

    (remhash square shadow)
    
    (when (zerop shadow-size)
      (return-from component-value 1))

    (dolist ( piece (rollable-pieces (opponent player)))
      (loop for square being the hash-keys of (get-shadow piece)
	    do
	    (remhash square shadow)))

    (/ (hash-table-count shadow)
	       shadow-size)))

(defclass escape-route ( strategy-component)
  (( name
     :initform "Escape Route")
   ( coef
     :initform '(  5.990918 22.666609))))

(defmethod cmp-pieces (( self escape-route) ( piece1 piece) ( piece2 piece))
  (let (( gf (godfather (player self))))
    (signum (- (dist piece1 gf) (dist piece2 gf)))))

(defmethod component-value (( self escape-route))
  (let* (( blo (blocker (player  self)))
	 ( the-player (player self))
	 ( gf (godfather the-player))
	 ( pieces (sort (copy-list (rollable-pieces the-player))
			#'(lambda ( p1 p2) (cmp-pieces self p1 p2))))
	 
	 ( closest (car pieces))
	 ( approach (approach-direction closest gf))
	 square )

    (loop for dir in (sweep-directions approach)
	  do
	  (setf square (get-square blo
				   (+ (y-coord (state gf)) (deltay dir))
				   (+ (x-coord (state gf)) (deltax dir))))
	  (when (and square
		     (null (occupant square)))
	    (return-from component-value 1)))
    0))
		     

(defclass gf-free-squares ( strategy-component )
  (( name
     :initform "Godfather Guarded")
   ( coef
     :initform '( 9.370575 15.605197))))

(defmethod component-value (( self gf-free-squares))
  (let* (( free 0)
	 ( edge 0)
	 ( occupied 0)
	 square
	 ( blo (blocker (player (strategy self))))
	 ( state (state (godfather (player (strategy self)))))

	 )

    (loop for dx in '( -1 1  0 0)
	  for dy in '(  0 0 -1 1)
	  do
	  (setf square (get-square blo
				   (+ dy (y-coord state))
				   (+ dx (x-coord state))))
	  (cond
	    ((not square)
	     (incf edge))

	    ((not (occupant square))
	     (incf free))

	    (t ;; (occupant square)
	     (incf occupied))))

    (cond
      ;; if gf is in the middle of the board and not touching any
      ;;    piece (of the same player
      ((= 4 free)
       0)
             
      ;; if gf is protected on all open/board sides
      ((= 4 (+ occupied edge))
       1)
        
      ;; if all board sides are open
      ((= 4 (+ free edge))
       0)

      (t
       (/ (- 4 free) 4)))))
