;; $Id: ui.lisp,v 1.6 2005/05/23 21:26:28 jimka Exp $

#|

 This software is Copyright (c) 2005 Jim Newton <jimka@rdrop.com>

 Jim Newton grants you the rights to distribute
 and use this software as governed by the terms
 of the Lisp Lesser GNU Public License
 (http://opensource.franz.com/preamble.html),
 known as the LLGPL.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
|#

(in-package blocker)

(defmethod visual-game (( blo blocker))
  (let* (( virtual (make-instance 'virtual-3d-equipment
				  :blocker blo))
	 ( ltk-blocker (make-instance 'ltk-blocker
				      :blocker blo
				      :virtual virtual)))
    (setf (event-handlers blo)
	  (list
	   virtual
	   ltk-blocker))))

(defun play-game ( &key verbose ( ltk t))
  (catch 'finish-blocker
    (let (( blo (make-instance 'blocker :verbose verbose)))
      (when ltk (visual-game blo))
      (do-event blo "rs")
      (play blo))))

(defun tk-game ( &key verbose ( ltk t))
  (catch 'finish-blocker
    (let (( blo (make-instance 'blocker :verbose verbose)))
      (when ltk (visual-game blo))
      (do-event blo "rs")
      (do-event blo "tw")
      (play blo))))

(defun self-play ( &key verbose)
  (catch 'finish-blocker
    (let (( blo (make-instance 'blocker :verbose verbose)))
      (visual-game blo) 
      (do-event blo "rs")    
      (do-event blo "cw")
      (do-event blo "cb")	      
      (play blo))))

(defun incrementally-improve-strategy ( &key verbose)
  (let (( blo (make-instance 'blocker :verbose verbose))
	hash
	comp
	strategy-white
	
	;; choose 0 or 1 randomly
	( index (random 2))

	;; choose a number between 0.5 and 2.0
	( rand (+ 0.5 (* 1.5 (random 1.0))))
	)
    
    (do-event blo "rs")  ;; read strategy file
    (do-event blo "cw")  ;; computer plays white
    (do-event blo "cb")	 ;; computer also plays black
    (setf strategy-white (strategy (white blo))
	  (remaining-moves blo) 300
	  ;; choose a strategy component at random from white
	  comp (nth (random (length (components strategy-white)))
		    (components strategy-white))

	  ;; adjust either the 1st order or 2nd order coeficient of
	  ;; the strategy component.  worst case is doubling it or
	  ;; halving it.
	  (nth index (coef comp)) (* rand (nth index (coef comp))))
    
    ;; does this perterbation of white's strategy cause white to win?
    ;; if not, then the perterbation was not successful.
    ;; thus return nil
    (unless (eql (white blo) (play blo))
      (format t "DID NOT FIND WINNING STRATEGY~%")
      (return-from incrementally-improve-strategy nil))

    (format t "NEW STRATEGY 50% verified~%")
    ;; save the strategy coefs to the hash table so we can retreieve it
    ;;  below after reallocating a new game structure.
    (setf hash (strat-to-hash (strategy (black blo)))
	  blo  (make-instance 'blocker :verbose verbose)
	  (remaining-moves blo) 300)
    
    (do-event blo "cw")  ;; computer plays white
    (do-event blo "cb")  ;; computer also plays black
    ;; retreive the strategy coefs from the hash-table
    (hash-to-strat (strategy (white blo)) hash)
    (swap-strategies blo)

    ;; does this perterbation of black's strategy cause black to win?
    ;; if not, then the perterbation was not successful.
    ;; thus return nil
    (unless (eql (black blo) (play blo))
      (format t "DID NOT FIND WINNING STRATEGY~%")
      (return-from incrementally-improve-strategy nil))

    ;; it we have succesfully found a better strategy then save
    ;; it.
    (format t "FOUND WINNING STRATEGY~%")
    (do-event blo "ws")

    t))

(defun improve-strategy ( &key verbose)
  (let (( game-number 1)
	( num-improvements  0))
    (loop do
	  (format t "==================================================~%")
	  (format t "   IMPROVEMENTS ~D/~D = ~F~%"
		  num-improvements game-number 
		  (* 100 (coerce (/ num-improvements game-number) 'float)))
	  (format t "==================================================~%")
	  (when (incrementally-improve-strategy :verbose verbose)
	    (incf num-improvements))
	  (incf game-number)
	  (clear-cache))))



;;   (profile:profile-all :package :blocker )
;; (profile:report-time)
;; 
;; (profile:unprofile)
;; (profile:profile-all :package :blocker :callers-p t)

